/*
 * Copyright (c) 2010-2011 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice,  this list of  conditions and the following disclaimer in
 *      the  documentation  and/or  other   materials provided  with  the
 *      distribution.
 *
 * THIS  SOFTWARE IS PROVIDED BY  THE  COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND  ANY  EXPRESS OR IMPLIED  WARRANTIES,  INCLUDING,  BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES  OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR  PURPOSE ARE DISCLAIMED. IN  NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR      CONTRIBUTORS  BE LIABLE FOR   ANY    DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF  SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE   OF THIS SOFTWARE, EVEN   IF ADVISED OF   THE POSSIBILITY OF SUCH
 * DAMAGE.
 *
 *                                      Anthony Mallet on Wed Dec  1 2010
 */

#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <err.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <sys/mman.h>
#ifdef __linux__
# include <linux/videodev2.h>
#endif
#ifdef __NetBSD__
# include <sys/videoio.h>
#endif

#include "viam/viam_acq_driver.h"


/* --- local data ---------------------------------------------------- */

typedef struct viamv4l2_camera_t {
  int fd;				/**< v4l2 device. */
  int input;				/**< input source number. */
  int mmap;				/**< mmap() (true) or read() (false) */
  double fps;				/**< desired fps */

  size_t isize;				/**< image size, bytes */
  size_t rowstride;			/**< line size, bytes */

  unsigned int nbuf;			/**< number of buffers */
  unsigned int cbuf;			/**< current buffer */
  struct viamv4l_buffer {
    char *data;
    size_t length;
  } *image;				/**< image buffers */
  struct timeval timestamp;		/**< last image timestamp */
  double toffset;			/**< internal timestamp offset */

					/**< camera features. */
  struct v4l2_queryctrl usrfeat[V4L2_CID_LASTP1-V4L2_CID_BASE];
#ifdef __linux__
  struct v4l2_queryctrl camfeat[16];
#endif

#define VIAM_V4L2_CTRL_DISABLED	(		\
    V4L2_CTRL_FLAG_DISABLED	|		\
    V4L2_CTRL_FLAG_GRABBED	|		\
    V4L2_CTRL_FLAG_READ_ONLY	|		\
    V4L2_CTRL_FLAG_INACTIVE			\
    )

} *viamv4l2_camera_t;

static int	viamv4l2_hardware_list(viam_handle_t vh);
static int	viamv4l2_hardware_attach(viam_handle_t vh,
			viam_camera_t camera);
static void	viamv4l2_hardware_detach(viam_handle_t vh,
			viam_camera_t camera);
static int	viamv4l2_camera_configure(viam_handle_t vh,
			viam_camera_t camera);
static int	viamv4l2_camera_unconfigure(viam_handle_t vh,
			viam_camera_t camera);
static int	viamv4l2_datatransmit(viam_handle_t vh, viam_camera_t camera,
			viam_onoff_t on);
static int	viamv4l2_dataflush(viam_handle_t vh, viam_camera_t camera);
static int	viamv4l2_capture(viam_handle_t vh, viam_camera_t camera);
static char *	viamv4l2_capturedata(viam_handle_t vh, viam_camera_t camera,
			size_t *bytes, size_t *rowstride);
static struct timeval
		viamv4l2_capturetimestamp(viam_handle_t vh,
			viam_camera_t camera);
static int	viamv4l2_capturedone(viam_handle_t vh, viam_camera_t camera);
static int	viamv4l2_queryhwmodes(viam_handle_t vh, viam_camera_t camera,
			int offset, viam_hwmode_t *supported, size_t n);
static int	viamv4l2_queryhwfeature(viam_handle_t vh, viam_camera_t camera,
			viam_filtertype_t filter, unsigned int *methods,
			unsigned int *aumodes);
static int	viamv4l2_sethwfeaturemode(viam_handle_t vh,
			viam_camera_t camera, viam_filtertype_t filter,
			viam_filtermethod_t method,
			viam_filterautomode_t aumode);
static int	viamv4l2_sethwfeaturevalue(viam_handle_t vh,
			viam_camera_t camera, viam_filtertype_t filter,
			const viam_filterargs_obj args);

static void	viamv4l2_camerainfo(const char *dev, int input,
			struct v4l2_capability *cap);

static int	viamv4l2_v4lfeature(viamv4l2_camera_t c,
			viam_filtertype_t filter, struct v4l2_queryctrl **info,
			struct v4l2_queryctrl **info_auto);
static int	viamv4l2_setfeat(viam_handle_t vh, int fd, unsigned int id,
			unsigned int value);
static int	viamv4l2_hw_fmt(viam_hwformat_t swfmt, uint32_t *hwfmt);
static int	viamv4l2_sw_fmt(uint32_t hwfmt, viam_hwformat_t *swfmt);
static int	viamv4l2_hw_fps(viam_hwfps_t swfps, double *hwfps);

const struct viam_acqdev_obj viam_acqdev_entry = {
  "v4l2",
  NULL,
  NULL,
  viamv4l2_hardware_list,
  viamv4l2_hardware_attach,
  viamv4l2_hardware_detach,
  viamv4l2_camera_configure,
  viamv4l2_camera_unconfigure,
  viamv4l2_datatransmit,
  viamv4l2_dataflush,
  viamv4l2_capture,
  viamv4l2_capturedata,
  viamv4l2_capturetimestamp,
  viamv4l2_capturedone,
  viamv4l2_queryhwmodes,
  viamv4l2_queryhwfeature,
  viamv4l2_sethwfeaturemode,
  viamv4l2_sethwfeaturevalue
};


/* --- viamv4l2_hardware_list ---------------------------------------- */

/** \brief Print available camera devices on all buses
 *
 * \param[in] vh	Library handle.
 *
 * \return 0 on success.
 */

static int
viamv4l2_hardware_list(viam_handle_t vh)
{
#define VIAM_MAX_DEVICE	10
  struct v4l2_capability cap;
  struct v4l2_input input;
  char device[64];
  struct stat sb;
  int fd;
  int i;

  /* iterate over /dev/video#i */
  for(i=0; i<VIAM_MAX_DEVICE; i++) {

    /* open device */
    snprintf(device, sizeof(device), "/dev/video%d", i);
    if (stat(device, &sb)) continue;
    fd = open(device, O_RDWR);
    if (fd < 0) {
      warnx("unable to access `%s': %s", device, strerror(errno));
      continue;
    }

    /* query capabilities */
    if (ioctl(fd, VIDIOC_QUERYCAP, &cap) < 0) {
      warnx("unable to query `%s': %s", device, strerror(errno));
      close(fd);
      continue;
    }
    if (!(cap.capabilities & V4L2_CAP_VIDEO_CAPTURE)) {
      close(fd);
      continue;
    }
    if (!(cap.capabilities & (V4L2_CAP_READWRITE|V4L2_CAP_STREAMING))) {
      close(fd);
      continue;
    }

    /* scan inputs */
    for(input.index = 0;;input.index++) {
      if (ioctl(fd, VIDIOC_ENUMINPUT, &input))
	break;

      if (input.type == V4L2_INPUT_TYPE_CAMERA)
	viamv4l2_camerainfo(device, input.index, &cap);
    }
    close(fd);
  }

  return 0;
}


/* --- viamv4l2_hardware_attach -------------------------------------- */

/** \brief Allocate camera ports and handles
 *
 * \param[in] vh	Library handle.
 * \param[in] camera	Camera handle to attach.
 *
 * \return 0 on success, or an error.
 */

static int
viamv4l2_hardware_attach(viam_handle_t vh, viam_camera_t camera)
{
  struct v4l2_capability cap;
  struct v4l2_input input;
  viamv4l2_camera_t c;
  struct timespec tsr, tsm;
  char device[1024];
  char *unit, *end;
  int fd, n;

  /* uid is "device path"[:"input source"] */
  strncpy(device, viam_camera_uid(camera), sizeof(device));
  unit = strchr(device, ':');
  if (unit) {
    *(unit++) = '\0';

    input.index = strtol(unit, &end, 0);
    if (end == unit || *end != '\0')
      return viam_seterrno(vh, Eviamlib_BAD_ID);
  }

  /* open device */
  fd = open(device, O_RDWR);
  if (fd < 0) {
    warnx("unable to access `%s': %s", device, strerror(errno));
    return viam_seterrno(vh, Eviamlib_BAD_ID);
  }

  /* check capabilities */
  if (ioctl(fd, VIDIOC_QUERYCAP, &cap) < 0) {
    warnx("unable to query `%s': %s", device, strerror(errno));
    goto badid;
  }
  if (!(cap.capabilities & V4L2_CAP_VIDEO_CAPTURE)) {
    warnx("`%s' is not a video capture device", device);
    goto badid;
  }
  if (!(cap.capabilities & (V4L2_CAP_READWRITE|V4L2_CAP_STREAMING))) {
    warnx("`%s' is not capable enough", device);
    goto badid;
  }

  /* auto-select input if not specified, or check input is valid */
  if (!unit) {
    for(input.index = 0;;input.index++) {
      input.type = -1;
      if (ioctl(fd, VIDIOC_ENUMINPUT, &input))
	break;
      if (input.type == V4L2_INPUT_TYPE_CAMERA)
	break;
    }
  } else {
    if (ioctl(fd, VIDIOC_ENUMINPUT, &input)) {
      input.index = -1;
    }
  }
  if (input.index == -1) {
    warnx("`%s' has no camera input", device);
    goto badid;
  }
  if (input.type != V4L2_INPUT_TYPE_CAMERA) {
      warnx("input #%d (%s) for `%s' is not a camera",
	    input.index, input.name, viam_camera_name(camera));
      goto badid;
  }

  if (ioctl(fd, VIDIOC_S_INPUT, &input) < 0) {
    warnx("`%s' has no camera input", device);
    goto badid;
  }

  /* register camera */
  c = malloc(sizeof(*c));
  if (!c) {
    warnx("unable to allocate memory for camera %s", viam_camera_name(camera));
    return viam_seterrno(vh, Eviamlib_MEMORY);
  }

  c->fd = fd;
  c->input = input.index;
  c->mmap = (cap.capabilities & V4L2_CAP_STREAMING)?1:0;
  c->fps = VIAM_HWFPS_INVALID;
  c->isize = 0;
  c->rowstride = 0;
  c->nbuf = 0;
  c->cbuf = 0;
  c->image = NULL;

  /* camera features */
  for(n = 0; n < sizeof(c->usrfeat)/sizeof(c->usrfeat[0]); n++) {
    memset(&c->usrfeat[n], 0, sizeof(c->usrfeat[n]));
    c->usrfeat[n].id = n + V4L2_CID_BASE;
    if (ioctl(fd, VIDIOC_QUERYCTRL, &c->usrfeat[n]) < 0)
      c->usrfeat[n].flags = V4L2_CTRL_FLAG_DISABLED;
  }
#ifdef __linux__
  for(n = 0; n < sizeof(c->camfeat)/sizeof(c->camfeat[0]); n++) {
    memset(&c->camfeat[n], 0, sizeof(c->camfeat[n]));
    c->camfeat[n].id = n + V4L2_CID_CAMERA_CLASS_BASE;
    if (ioctl(fd, VIDIOC_QUERYCTRL, &c->camfeat[n]) < 0)
      c->camfeat[n].flags = V4L2_CTRL_FLAG_DISABLED;
  }
#endif

  /* offset between CLOCK_MONOTONIC and CLOCK_REALTIME */
  clock_gettime(CLOCK_REALTIME, &tsr);
  clock_gettime(CLOCK_MONOTONIC, &tsm);
  c->toffset = tsr.tv_sec - tsm.tv_sec + (tsr.tv_nsec - tsm.tv_nsec)*1e-9;

  viam_camera_sethwdata(camera, c);

  warnx("camera `%s' is %s (%s)",
	viam_camera_name(camera), cap.card, cap.driver);
  warnx("using input #%d (%s) for camera `%s'",
	input.index, input.name, viam_camera_name(camera));
  warnx("using %s performance capture for camera `%s'",
	c->mmap?"high":"low", viam_camera_name(camera));
  return 0;

badid:
  close(fd);
  return viam_seterrno(vh, Eviamlib_BAD_ID);
}


/* --- viamv4l2_hardware_detach -------------------------------------- */

/** \brief Cleanup dcv4l2 handle
 *
 * \param[in] vh	Library handle.
 * \param[in] camera	Camera handle.
 */

static void
viamv4l2_hardware_detach(viam_handle_t vh, viam_camera_t camera)
{
  viamv4l2_camera_t hw;
  int i;

  hw = viam_camera_hwdata(camera);
  if (hw) {
    viam_camera_sethwdata(camera, NULL);
    close(hw->fd);
    if (hw->image) {
      for(i = 0; i < hw->nbuf; i++) {
	if (hw->image[i].data) {
	  if (hw->mmap)
	    munmap(hw->image[i].data, hw->image[i].length);
	  else
	    free(hw->image[i].data);
	}
      }
      free(hw->image);
    }
    free(hw);
  }
}


/* --- viamv4l2_camera_configure ------------------------------------- */

/** \brief Setup camera with current configuration parameters.
 *
 * \param[in] vh	Library handle.
 * \param[in] camera	Camera handle.
 *
 * \return 0 on success
 */

static int
viamv4l2_camera_configure(viam_handle_t vh, viam_camera_t camera)
{
  viamv4l2_camera_t c = viam_camera_hwdata(camera);
  struct v4l2_requestbuffers reqbuf;
  struct v4l2_format hwfmt;
  struct v4l2_buffer buffer;
  viam_hwmode_t swmode;
  int width, height;
  int i;

  if (!c) return viam_seterrno(vh, Eviamlib_NO_SUCH_CAMERA);

  /* configure hardware mode */
  if (!viam_camera_hwmode(vh, camera))
    return viam_seterrno(vh, Eviamlib_UNSUPPORTED_HWMODE);
  swmode = *viam_camera_hwmode(vh, camera);

  if (viamv4l2_hw_fps(swmode.fps, &c->fps))
    return viam_seterrno(vh, Eviamlib_UNSUPPORTED_HWMODE);
  if (swmode.trigger != VIAM_HWTRIGGER_INTERNAL)
    return viam_seterrno(vh, Eviamlib_UNSUPPORTED_HWMODE);
  if (viamv4l2_hw_fmt(swmode.format, &hwfmt.fmt.pix.pixelformat))
    return viam_seterrno(vh, Eviamlib_UNSUPPORTED_HWMODE);
  if (viam_hwmode_size(vh, &swmode, &width, &height, NULL, NULL))
    return viam_seterrno(vh, Eviamlib_UNSUPPORTED_HWMODE);
  hwfmt.fmt.pix.width = width;
  hwfmt.fmt.pix.height = height;

  hwfmt.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
  if (ioctl(c->fd, VIDIOC_S_FMT, &hwfmt))
    return viam_seterrno(vh, Eviamlib_UNSUPPORTED_HWMODE);
  if (hwfmt.fmt.pix.width != width || hwfmt.fmt.pix.height != height)
    return viam_seterrno(vh, Eviamlib_UNSUPPORTED_HWMODE);

  switch(c->mmap) {
    case 0:
      /* alloc 1 image buffer for read() */
      c->image = malloc(sizeof(*c->image));
      if (!c->image) return viam_seterrno(vh, Eviamlib_MEMORY);
      c->image[0].data = malloc(hwfmt.fmt.pix.sizeimage);
      if (!c->image[0].data) return viam_seterrno(vh, Eviamlib_MEMORY);
      c->image[0].length = hwfmt.fmt.pix.sizeimage;
      c->nbuf = 1;
      break;

    default:
      /* alloc 2 image buffers for mmap() */
      memset(&reqbuf, 0, sizeof(reqbuf));
      reqbuf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
      reqbuf.memory = V4L2_MEMORY_MMAP;
      reqbuf.count = 2;
      if (ioctl(c->fd, VIDIOC_REQBUFS, &reqbuf) < 0) {
	warnx("camera %s cannot request buffers: %s",
	      viam_camera_name(camera), strerror(errno));
	return viam_seterrno(vh, Eviamlib_FWBUS);
      }

      c->image = malloc(reqbuf.count * sizeof(*c->image));
      if (!c->image) return viam_seterrno(vh, Eviamlib_MEMORY);

      for (i = 0; i < reqbuf.count; i++) {
        memset(&buffer, 0, sizeof(buffer));
        buffer.type = reqbuf.type;
	buffer.memory = V4L2_MEMORY_MMAP;
        buffer.index = i;
        if (ioctl(c->fd, VIDIOC_QUERYBUF, &buffer) < 0) {
	  warnx("camera %s cannot query buffers: %s",
		viam_camera_name(camera), strerror(errno));
	  return viam_seterrno(vh, Eviamlib_FWBUS);
	}

        c->image[i].data = mmap(NULL, buffer.length,
				PROT_READ | PROT_WRITE,	/* recommended */
				MAP_SHARED,		/* recommended */
				c->fd, buffer.m.offset);
        if (c->image[i].data == MAP_FAILED) {
	  /* should unmap() the buffers mapped so far. */
	  warnx("camera %s cannot map buffers: %s",
		viam_camera_name(camera), strerror(errno));
	  return viam_seterrno(vh, Eviamlib_MEMORY);
        }
        c->image[i].length = buffer.length; /* remember for munmap() */
      }

      c->nbuf = reqbuf.count;
      break;
  }

  c->isize = hwfmt.fmt.pix.sizeimage;
  c->rowstride = hwfmt.fmt.pix.bytesperline;
  return 0;
}


/* --- viamv4l2_camera_unconfigure ----------------------------------- */

/** \brief Reset bank configuration
 *
 * \param[in] vh	Library handle.
 * \param[in] camera	Camera handle.
 *
 * \return 0 on success
 */

static int
viamv4l2_camera_unconfigure(viam_handle_t vh, viam_camera_t camera)
{
  viamv4l2_camera_t c = viam_camera_hwdata(camera);
  int i;

  if (!c) return viam_seterrno(vh, Eviamlib_NO_SUCH_CAMERA);
  if (c->image) {
    for(i = 0; i < c->nbuf; i++) {
      if (c->mmap)
        munmap(c->image[i].data, c->image[i].length);
      else
	free(c->image[i].data);
    }
    free(c->image);
    c->image = NULL;
    c->nbuf = 0;
  }
  c->isize = 0;
  c->rowstride = 0;

  return 0;
}


/* --- viam_datatransmit --------------------------------------------- */

int
viamv4l2_datatransmit(viam_handle_t vh, viam_camera_t camera, viam_onoff_t on)
{
  viamv4l2_camera_t c = viam_camera_hwdata(camera);
  struct v4l2_buffer buffer;
  int arg = (on == VIAM_ON)?VIDIOC_STREAMON:VIDIOC_STREAMOFF;
  int i;

  if (!c) return viam_seterrno(vh, Eviamlib_NOT_CONFIGURED);
  if (!c->mmap) return 0 /* nothing to do */;

  for(i=0; i<c->nbuf; i++) {
    memset(&buffer, 0, sizeof(buffer));
    buffer.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    buffer.memory = V4L2_MEMORY_MMAP;
    buffer.index = i;
    if (ioctl(c->fd, VIDIOC_QUERYBUF, &buffer) < 0) {
      warnx("camera %s cannot query buffer: %s",
	    viam_camera_name(camera), strerror(errno));
    }
    if (on == VIAM_ON && (buffer.flags & V4L2_BUF_FLAG_QUEUED))
      continue;
    if (on == VIAM_OFF &&
	!(buffer.flags & (V4L2_BUF_FLAG_QUEUED|V4L2_BUF_FLAG_DONE)))
      continue;

    if (ioctl(c->fd, (on == VIAM_ON)?VIDIOC_QBUF:VIDIOC_DQBUF, &buffer) < 0) {
      warnx("camera %s cannot %squeue buffer: %s",
	    viam_camera_name(camera), (on == VIAM_ON)?"":"de", strerror(errno));
    }
  }

  i = V4L2_BUF_TYPE_VIDEO_CAPTURE;
  if (ioctl(c->fd, arg, &i) < 0)
    return viam_seterrno(vh, Eviamlib_FWBUS);

  c->cbuf = 0;
  return 0;
}


/* --- viam_dataflush ------------------------------------------------ */

static int
viamv4l2_dataflush(viam_handle_t vh, viam_camera_t camera)
{
  return 0;
}


/* --- viamv4l2_capture ---------------------------------------------- */

static int
viamv4l2_capture(viam_handle_t vh, viam_camera_t camera)
{
  viamv4l2_camera_t c = viam_camera_hwdata(camera);
  struct v4l2_buffer buffer;
  struct timespec now;
  double delay;
  ssize_t s;
  if (!c || !c->image) return viam_seterrno(vh, Eviamlib_NOT_CONFIGURED);

  /* honor framerate: skip frames */
  while(1) {
    clock_gettime(CLOCK_REALTIME, &now);
    delay = now.tv_sec + now.tv_nsec * 1e-9;
    delay -= c->timestamp.tv_sec + c->timestamp.tv_usec * 1e-6;
    if (delay > 0. && c->fps > 0. && delay < 0.9/c->fps) {
      buffer.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
      buffer.memory = V4L2_MEMORY_MMAP;
      buffer.index = c->cbuf;
      if (ioctl(c->fd, VIDIOC_DQBUF, &buffer) < 0) {
	warnx("camera %s cannot dequeue buffer: %s",
	      viam_camera_name(camera), strerror(errno));
	return viam_seterrno(vh, Eviamlib_FWBUS);
      }
      if (ioctl(c->fd, VIDIOC_QBUF, &buffer) < 0) {
	warnx("camera %s cannot queue buffer: %s",
	      viam_camera_name(camera), strerror(errno));
      }
      c->cbuf = (c->cbuf+1) % c->nbuf;
    } else break;
  }

  switch(c->mmap) {
    case 0:
      s = read(c->fd, c->image[0].data, c->isize);
      if (s != c->isize) {
	if (s < 0)
	  warnx("camera %s: %s", viam_camera_name(camera), strerror(errno));
	return viam_seterrno(vh, Eviamlib_FWBUS);
      }
      break;

    default:
      memset(&buffer, 0, sizeof(buffer));
      buffer.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
      buffer.memory = V4L2_MEMORY_MMAP;
      buffer.index = c->cbuf;
      if (ioctl(c->fd, VIDIOC_DQBUF, &buffer) < 0) {
	warnx("camera %s cannot dequeue buffer: %s",
	      viam_camera_name(camera), strerror(errno));
	return viam_seterrno(vh, Eviamlib_FWBUS);
      }
      delay = buffer.timestamp.tv_sec + buffer.timestamp.tv_usec/1e6;
      if (now.tv_sec - delay > 10)
	delay+= c->toffset;

      c->timestamp.tv_sec = delay;
      c->timestamp.tv_usec = (delay - c->timestamp.tv_sec) * 1000000;
      break;
  }

  return 0;
}


/* --- viamv4l2_capturedata ------------------------------------------ */

static char *
viamv4l2_capturedata(viam_handle_t vh, viam_camera_t camera,
		     size_t *bytes, size_t *rowstride)
{
  viamv4l2_camera_t c = viam_camera_hwdata(camera);

  if (!c || !c->image) {
    viam_seterrno(vh, Eviamlib_NOT_CONFIGURED);
    return NULL;
  }

  if (bytes) *bytes = c->isize;
  if (rowstride) *rowstride = c->rowstride;
  return c->image[c->cbuf].data;
}


/* --- viamv4l2_capturetimestamp ------------------------------------- */

static struct timeval
viamv4l2_capturetimestamp(viam_handle_t vh, viam_camera_t camera)
{
  viamv4l2_camera_t c = viam_camera_hwdata(camera);

  if (!c || !c->image) {
    viam_seterrno(vh, Eviamlib_NOT_CONFIGURED);
    return (struct timeval){ 0 };
  }

  return c->timestamp;
}


/* --- viamv4l2_capturedone ------------------------------------------ */

static int
viamv4l2_capturedone(viam_handle_t vh, viam_camera_t camera)
{
  viamv4l2_camera_t c = viam_camera_hwdata(camera);
  struct v4l2_buffer buffer;

  switch(c->mmap) {
    case 0:
      /* nothing to do */
      break;

    default:
      memset(&buffer, 0, sizeof(buffer));
      buffer.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
      buffer.memory = V4L2_MEMORY_MMAP;
      buffer.index = c->cbuf;
      if (ioctl(c->fd, VIDIOC_QBUF, &buffer) < 0) {
	warnx("camera %s cannot queue buffer: %s",
	      viam_camera_name(camera), strerror(errno));
      }

      c->cbuf = (c->cbuf+1) % c->nbuf;
      break;
  }

  return 0;
}


/* --- viamv4l2_queryhwmodes ----------------------------------------- */

static int
viamv4l2_queryhwmodes(viam_handle_t vh, viam_camera_t camera, int offset,
		      viam_hwmode_t *supported, size_t n)
{
  viamv4l2_camera_t c = viam_camera_hwdata(camera);
  struct v4l2_fmtdesc fmt;
  struct v4l2_format pix;
  viam_hwmode_t swmode;
  int width, height;
  int i, count;

  if (!c) { viam_seterrno(vh, Eviamlib_NOT_CONFIGURED); return 0; }

  /* no crop */
  swmode.crop = VIAM_HW_FIXED;
  /* fps is free */
  swmode.fps = VIAM_HWFPS_FREE;
  /* internal trigger is the only trigger format */
  swmode.trigger = VIAM_HWTRIGGER_INTERNAL;

  count = 0;
  for(fmt.index = 0;; fmt.index++) {
    fmt.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    if (ioctl(c->fd, VIDIOC_ENUM_FMT, &fmt)) break;
    if (viamv4l2_sw_fmt(fmt.pixelformat, &swmode.format)) continue;

    /* get available sizes */
    for(i = 0; i < viam_hwsize_vals; i++) {
      swmode.size = i;
      if (viam_hwmode_size(vh, &swmode, &width, &height, NULL, NULL))
	continue;

      pix.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
      pix.fmt.pix.pixelformat = fmt.pixelformat;
      pix.fmt.pix.width = width;
      pix.fmt.pix.height = height;
      if (ioctl(c->fd, VIDIOC_S_FMT, &pix)) continue;
      if (pix.fmt.pix.width != width || pix.fmt.pix.height != height) continue;

      if (count >= offset && (count-offset)<(signed)n)
	supported[count-offset] = swmode;
      count++;
    }
  }

  return (count>=offset)?(count-offset>n?n:count-offset):0;
}


/* --- viamv4l2_queryhwfeature --------------------------------------- */

static int
viamv4l2_queryhwfeature(viam_handle_t vh,
			viam_camera_t camera, viam_filtertype_t filter,
			unsigned int *methods, unsigned int *aumodes)
{
  viamv4l2_camera_t c = viam_camera_hwdata(camera);
  struct v4l2_queryctrl *info, *info_auto;
  int s;

  if (methods) *methods = 0;
  if (aumodes) *aumodes = 0;

  if (!c) return viam_seterrno(vh, Eviamlib_NOT_CONFIGURED);

  s = viamv4l2_v4lfeature(c, filter, &info, &info_auto);
  if (s || !info) return 0;

  if (!(info->flags & VIAM_V4L2_CTRL_DISABLED)) {
    if (methods) *methods |= VIAM_FILTER_HARDWARE;
    if (aumodes) *aumodes |= VIAM_FILTER_MANUAL;

    if (info->type & V4L2_CTRL_TYPE_BOOLEAN) {
      if (methods) *methods |= VIAM_FILTER_OFF;
    }
  }

  if (info_auto && !(info_auto->flags & VIAM_V4L2_CTRL_DISABLED)) {
    if (methods) *methods |= VIAM_FILTER_HARDWARE;
    if (aumodes) *aumodes |= VIAM_FILTER_HARDWARE_AUTO;

    if (info_auto->type & V4L2_CTRL_TYPE_BOOLEAN) {
      if (methods) *methods |= VIAM_FILTER_OFF;
    }
  }

  /* special case */
  if (filter == VIAM_FILTER_WHITE_BALANCE) {
    info = &c->usrfeat[V4L2_CID_DO_WHITE_BALANCE-V4L2_CID_BASE];
    if (!(info->flags & VIAM_V4L2_CTRL_DISABLED)) {
      if (aumodes) *aumodes |= VIAM_FILTER_HARDWARE_ONE_PUSH;
    }
  }
  return 0;
}


/* --- viamv4l2_sethwfeaturemode ------------------------------------- */

/** \brief Set hardware feature mode for a camera.
 *
 * \param[in] vh	Library handle.
 * \param[in] camera	Camera handle.
 * \param[in] filter	Feature type.
 * \param[in] method	Hardware mode.
 * \param[in] aumode	Auto mode.
 *
 * \return 0 on success.
 */

static int
viamv4l2_sethwfeaturemode(viam_handle_t vh, viam_camera_t camera,
			  viam_filtertype_t filter, viam_filtermethod_t method,
			  viam_filterautomode_t aumode)
{
  viamv4l2_camera_t c = viam_camera_hwdata(camera);
  struct v4l2_queryctrl *info, *info_auto;
  int s;

  if (!c) return viam_seterrno(vh, Eviamlib_NOT_CONFIGURED);

  s = viamv4l2_v4lfeature(c, filter, &info, &info_auto);
  if (s) switch(method) {
    case VIAM_FILTER_SOFTWARE:
    case VIAM_FILTER_OFF:
      return 0;

    default: return viam_seterrno(vh, s);
  }

  switch(method) {
    case VIAM_FILTER_SOFTWARE:
    case VIAM_FILTER_OFF:
      /* switch feature off if possible */
      if (info && !(info->flags & VIAM_V4L2_CTRL_DISABLED)) {
	if (info->type & V4L2_CTRL_TYPE_BOOLEAN) {
	  if (viamv4l2_setfeat(vh, c->fd, info->id, 0))
	    return viam_errno(vh);
	}
      }
      /* switch auto mode off if possible */
      if (info_auto && !(info_auto->flags & VIAM_V4L2_CTRL_DISABLED)) {
	if (info_auto->type & V4L2_CTRL_TYPE_BOOLEAN) {
	  if (viamv4l2_setfeat(vh, c->fd, info_auto->id, 0))
	    return viam_errno(vh);
	}
      }
      break;

    case VIAM_FILTER_HARDWARE:
      switch(aumode) {
	case VIAM_FILTER_MANUAL:
	case VIAM_FILTER_SOFTWARE_AUTO:
	case VIAM_FILTER_SOFTWARE_ONE_PUSH:
	  /* switch feature on if possible */
	  if (info && !(info->flags & VIAM_V4L2_CTRL_DISABLED)) {
	    if (info->type & V4L2_CTRL_TYPE_BOOLEAN) {
	      if (viamv4l2_setfeat(vh, c->fd, info->id, 1))
		return viam_errno(vh);
	    }
	  }
	  /* switch auto mode off if possible */
	  if (info_auto && !(info_auto->flags & VIAM_V4L2_CTRL_DISABLED)) {
	    if (info_auto->type & V4L2_CTRL_TYPE_BOOLEAN) {
	      if (viamv4l2_setfeat(vh, c->fd, info_auto->id, 0))
		return viam_errno(vh);
	    }
	  }
	  break;

	case VIAM_FILTER_HARDWARE_AUTO:
	  /* switch feature on if possible */
	  if (info && !(info->flags & VIAM_V4L2_CTRL_DISABLED)) {
	    if (info->type & V4L2_CTRL_TYPE_BOOLEAN) {
	      if (viamv4l2_setfeat(vh, c->fd, info->id, 1))
		return viam_errno(vh);
	    }
	  }
	  /* switch auto mode on */
	  if (info_auto && !(info_auto->flags & VIAM_V4L2_CTRL_DISABLED)) {
	    if (info_auto->type & V4L2_CTRL_TYPE_BOOLEAN) {
	      if (viamv4l2_setfeat(vh, c->fd, info_auto->id, 1))
		return viam_errno(vh);
	    } else
	      return viam_seterrno(vh, Eviamlib_NOT_IMPLEMENTED);
	  } else
	    return viam_seterrno(vh, Eviamlib_NOT_IMPLEMENTED);
	  break;

	case VIAM_FILTER_HARDWARE_ONE_PUSH:
	  if (filter == VIAM_FILTER_WHITE_BALANCE) {
	    info = &c->usrfeat[V4L2_CID_DO_WHITE_BALANCE-V4L2_CID_BASE];
	    if (!(info->flags & VIAM_V4L2_CTRL_DISABLED)) {
	      if (viamv4l2_setfeat(vh, c->fd, V4L2_CID_DO_WHITE_BALANCE, 1))
		return viam_errno(vh);
	    }
	    break;
	  }
	  /*FALLTHROUGH*/
	case VIAM_FILTER_AUTOINVALID:
	  return viam_seterrno(vh, Eviamlib_NOT_IMPLEMENTED);
      }
      break;

    case VIAM_FILTER_INVALID:
      return viam_seterrno(vh, Eviamlib_NOT_IMPLEMENTED);
  }

  return 0;
}


/* --- viamv4l2_sethwfeaturevalue ------------------------------------ */

/** \brief Set harware feature value for a camera.
 *
 * \param[in] vh	Library handle.
 * \param[in] camera	Camera handle.
 * \param[in] filter	Feature type.
 * \param[in] args	Feature value.
 *
 * \return 0 on success.
 */

static int
viamv4l2_sethwfeaturevalue(viam_handle_t vh, viam_camera_t camera,
			   viam_filtertype_t filter,
			   const viam_filterargs_obj args)
{
  viamv4l2_camera_t c = viam_camera_hwdata(camera);
  struct v4l2_queryctrl *info;
  uint32_t value;
  int s;

  if (!c) return viam_seterrno(vh, Eviamlib_NOT_CONFIGURED);

  s = viamv4l2_v4lfeature(c, filter, &info, NULL);
  if (s || !info) return viam_seterrno(vh, s);

  switch(filter) {
    case VIAM_FILTER_WHITE_BALANCE:
      value = args->red*info->maximum + (1.0 - args->red)*info->minimum;
      if (info->default_value != value) {
	if (viamv4l2_setfeat(vh, c->fd, info->id, value))
	  return viam_errno(vh);
	if (viamv4l2_setfeat(
	      vh, c->fd, V4L2_CID_BLUE_BALANCE,
	      args->blue*info->maximum + (1.0 - args->blue)*info->minimum))
	  return viam_errno(vh);
	info->default_value = value;
      }
      break;

    default:
      value = args->value*info->maximum + (1.0 - args->value)*info->minimum;
      if (info->default_value != value) {
	if (viamv4l2_setfeat(vh, c->fd, info->id, value))
	  return viam_errno(vh);
	info->default_value = value;
      }
      break;
  }

  return 0;
}


/* --- viamv4l2_camerainfo ------------------------------------------- */

static void
viamv4l2_camerainfo(const char *dev, int input, struct v4l2_capability *cap)
{
  printf("[node %s:%s:%d] %s (%s)\n",
	 viam_acqdev_entry.name, dev, input, cap->card, cap->driver);
}


/* --- viamv4l2_v4lfeature -------------------------------------------- */

static const struct {
  unsigned int v4lfeat;
  unsigned int v4lfeat_auto;
} featuremap[] = {
#ifdef __linux__
  [VIAM_FILTER_EXPOSURE] =	{  V4L2_CID_EXPOSURE,	V4L2_CID_EXPOSURE_AUTO },
  [VIAM_FILTER_ZOOM] =		{  V4L2_CID_ZOOM_ABSOLUTE,	-1 },
  [VIAM_FILTER_FOCUS] =		{  V4L2_CID_FOCUS_ABSOLUTE, V4L2_CID_FOCUS_AUTO },
#endif
  [VIAM_FILTER_NONE] =		{  -1,	-1 },
  [VIAM_FILTER_FORMAT] =	{  -1,	-1 },
  [VIAM_FILTER_SUBSAMPLE] =	{  -1,	-1 },
  [VIAM_FILTER_SCALE] =		{  -1,	-1 },
  [VIAM_FILTER_BRIGHTNESS] =	{  V4L2_CID_BRIGHTNESS,	-1 },
  [VIAM_FILTER_SHARPNESS] =	{  V4L2_CID_SHARPNESS,	-1 },
  [VIAM_FILTER_GAIN] =		{  V4L2_CID_GAIN,	V4L2_CID_AUTOGAIN },
  [VIAM_FILTER_SHUTTER] =	{  -1,	-1 },
  [VIAM_FILTER_HUE] =		{  V4L2_CID_HUE,	V4L2_CID_HUE_AUTO },
  [VIAM_FILTER_SATURATION] =	{  V4L2_CID_SATURATION,	-1 },
  [VIAM_FILTER_GAMMA] =		{  V4L2_CID_GAMMA,	-1 },
  [VIAM_FILTER_TEMPERATURE] =	{  V4L2_CID_WHITE_BALANCE_TEMPERATURE,	-1 },
#ifdef VIAM_OPENCV_SUPPORT
  [VIAM_FILTER_RECTIFY] =	{  -1,	-1 },
  [VIAM_FILTER_RECTIFY_NN] =	{  -1,	-1 },
  [VIAM_FILTER_DISTORT] =	{  -1,	-1 },
  [VIAM_FILTER_DISTORT_NN] =	{  -1,	-1 },
  [VIAM_FILTER_WHITE_BALANCE] =	{  V4L2_CID_RED_BALANCE, V4L2_CID_AUTO_WHITE_BALANCE },
  [VIAM_FILTER_LOG] =		{  -1,	-1 }
#endif
};

static int
viamv4l2_v4lfeature(viamv4l2_camera_t c, viam_filtertype_t filter,
		    struct v4l2_queryctrl **info,
		    struct v4l2_queryctrl **info_auto)
{
  unsigned int f;

  if (info) {
    f = featuremap[filter].v4lfeat;
    if (f == -1)
      *info = NULL;
    else if (f >= V4L2_CID_BASE && f < V4L2_CID_LASTP1)
      *info = &c->usrfeat[f - V4L2_CID_BASE];
#ifdef __linux__
    else if (f >= V4L2_CID_CAMERA_CLASS_BASE && f < V4L2_CID_PRIVACY)
      *info = &c->camfeat[f - V4L2_CID_CAMERA_CLASS_BASE];
#endif
    else
      *info = NULL;
  }

  if (info_auto) {
    f = featuremap[filter].v4lfeat_auto;
    if (f == -1)
      *info_auto = NULL;
    else if (f >= V4L2_CID_BASE && f < V4L2_CID_LASTP1)
      *info_auto = &c->usrfeat[f - V4L2_CID_BASE];
#ifdef __linux__
    else if (f >= V4L2_CID_CAMERA_CLASS_BASE && f < V4L2_CID_PRIVACY)
      *info_auto = &c->camfeat[f - V4L2_CID_CAMERA_CLASS_BASE];
#endif
    else
      *info_auto = NULL;
  }

  return *info == NULL ? Eviamlib_NOT_IMPLEMENTED : 0;
}

static int
viamv4l2_setfeat(viam_handle_t vh, int fd, unsigned int id, unsigned int value)
{
  struct v4l2_control ctrl;

  ctrl.id = id;
  ctrl.value = value;
  if (ioctl(fd, VIDIOC_S_CTRL, &ctrl)) {
    warnx("cannot set control: %s", strerror(errno));
    return viam_seterrno(vh, Eviamlib_FWBUS);
  }

  return 0;
}


/* --- viamv4l2_fmt -------------------------------------------------------- */

static const struct {
  viam_hwformat_t sw;
  uint32_t hw;
} fmtmap[] = {
#ifdef __linux__
  { VIAM_HWFMT_MONO16,		V4L2_PIX_FMT_Y16 },
  { VIAM_HWFMT_YVU422_VYUY,	V4L2_PIX_FMT_VYUY },
  { VIAM_HWFMT_YVU422_YVYU,	V4L2_PIX_FMT_YVYU },
  { VIAM_HWFMT_YUV444,		V4L2_PIX_FMT_YUV444 },
#endif
  { VIAM_HWFMT_MONO8,		V4L2_PIX_FMT_GREY },
  { VIAM_HWFMT_YUV422_UYVY,	V4L2_PIX_FMT_UYVY },
  { VIAM_HWFMT_YUV422_YUYV,	V4L2_PIX_FMT_YUYV },
  { VIAM_HWFMT_YUV422_YYUV,	V4L2_PIX_FMT_YYUV },
  { VIAM_HWFMT_RGB888,		V4L2_PIX_FMT_RGB24 }
};

static int
viamv4l2_hw_fmt(viam_hwformat_t swfmt, uint32_t *hwfmt)
{
  int i;

  for(i=0; i<sizeof(fmtmap)/sizeof(fmtmap[0]); i++)
    if (fmtmap[i].sw == swfmt) {
      *hwfmt = fmtmap[i].hw;
      return 0;
    }
  return Eviamlib_UNSUPPORTED_HWMODE;
}

static int
viamv4l2_sw_fmt(uint32_t hwfmt, viam_hwformat_t *swfmt)
{
  int i;

  for(i=0; i<sizeof(fmtmap)/sizeof(fmtmap[0]); i++)
    if (fmtmap[i].hw == hwfmt) {
      *swfmt = fmtmap[i].sw;
      return 0;
    }
  return Eviamlib_UNSUPPORTED_HWMODE;
}


/* --- viamv4l2_fps -------------------------------------------------------- */

static const struct {
  viam_hwfps_t sw;
  double hw;
} fpsmap[] = {
  { VIAM_HWFPS_1_875,	1.875},
  { VIAM_HWFPS_3_75,	3.75},
  { VIAM_HWFPS_7_5,	7.5},
  { VIAM_HWFPS_15,	15.},
  { VIAM_HWFPS_30,	30.},
  { VIAM_HWFPS_60,	60.},
  { VIAM_HWFPS_120,	120.},
  { VIAM_HWFPS_240,	240.},
  { VIAM_HWFPS_FREE,	0.}
};

static int
viamv4l2_hw_fps(viam_hwfps_t swfps, double *hwfps)
{
  int i;

  for(i=0; i<sizeof(fpsmap)/sizeof(fpsmap[0]); i++)
    if (fpsmap[i].sw == swfps) {
      *hwfps = fpsmap[i].hw;
      return 0;
    }
  return Eviamlib_UNSUPPORTED_HWMODE;
}
